/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file cli.c
* @brief Console output
* @author AlannY
* @date 2007-10-22
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>

#include <dbug.h>
#include <getopt.h>

#include "i18n.h"

#include "digest.h"
#include "entry.h"
#include "ff/crc32.h"
#include "ff/md4.h"
#include "ff/md5.h"
#include "ff/sha1.h"
#include "ff/sha256.h"
#include "ff/sha512.h"
#include "ff/ed2k.h"
#include "ff/tiger.h"

#include "cli.h"

inline void usage(void) {
    printf(_("Usage: readhash [options] FILE\n"
            "\n"
            "Options:\n"
            "        -h    --help        This help\n"
            "        -c    --crc32       Compute CRC32 hashsum\n"
            "        -M    --md4         Compute MD4 hashsum\n"
            "        -m    --md5         Compute MD5 hashsum\n"
            "        -s    --sha1        Compute SHA1 hashsum\n"
            "        -S    --sha256      Compute SHA256 hashsum\n"
            "        -a    --sha384      Compute SHA384 hashsum\n"
            "        -A    --sha512      Compute SHA512 hashsum\n"
            "        -t    --tiger       Compute Tiger hashsum\n"
            "        -e    --ed2k        Compute ED2K hashsum\n"
            "        -l    --ed2k-link   Compute ED2K link hashsum\n"
            "\n"
            "License: GPL\n"
            "Author: AlannY\n"));
    exit(EXIT_SUCCESS);
}

uint8_t cli(int *argc, char **argv[]) {
    
    struct option long_options[] = {
        {"help",     no_argument, NULL, 'h'},
        {"crc32",    required_argument, NULL, 'c'},
        {"md4",      required_argument, NULL, 'M'},
        {"md5",      required_argument, NULL, 'm'},
        {"sha1",     required_argument, NULL, 's'},
        {"sha256",   required_argument, NULL, 'S'},
        {"sha384",   required_argument, NULL, 'a'},
        {"sha512",   required_argument, NULL, 'A'},
        {"tiger",    required_argument, NULL, 't'},
        {"ed2k",     required_argument, NULL, 'e'},
        {"ed2k-link",required_argument, NULL, 'l'},
        {0, 0, 0, 0}
    };

    while( 1 ) {
        int32_t option_index = 0;
        int32_t c = getopt_long(*argc, *argv, "hc:M:m:s:S:a:A:t:e:l:", long_options, &option_index);
     
        if( c == -1 )
             break;

        uint8_t *filename = (uint8_t*)optarg;

#ifdef USE_GNOMEVFS
        if( strncmp((char*)filename, "file://", 7) != 0 ) {
            uint8_t *dir = (uint8_t*)malloc(sizeof(char)*4097);
            if( filename[0] != '/' ) {
                getcwd((char*)dir, 4096);
                strcat((char*)dir, "/");
            } else {
                dir[0] = '\0';
            }
            uint8_t *fn = (uint8_t*)malloc(sizeof(uint8_t*)*(strlen((char*)filename)+strlen((char*)dir)+8));
            strcpy((char*)fn, "file://");
            strcat((char*)fn, (char*)dir);
            strcat((char*)fn, (char*)filename);
            filename = fn;
        }
#endif
        
        switch( c ) {
            case 'c': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*9);
                    memset(digest, 0, sizeof(uint8_t)*9);
                    if( crc32_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'M': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*33);
                    memset(digest, 0, sizeof(uint8_t)*33);
                    if( md4_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'm': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*33);
                    memset(digest, 0, sizeof(uint8_t)*33);
                    if( md5_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 's': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*41);
                    memset(digest, 0, sizeof(uint8_t)*41);
                    if( sha1_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'S': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*65);
                    memset(digest, 0, sizeof(uint8_t)*65);
                    if( sha256_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'a': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*97);
                    memset(digest, 0, sizeof(uint8_t)*97);
                    if( sha384_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'A': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*129);
                    memset(digest, 0, sizeof(uint8_t)*129);
                    if( sha512_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 't': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*49);
                    memset(digest, 0, sizeof(uint8_t)*49);
                    if( tiger_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'e': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*33);
                    memset(digest, 0, sizeof(uint8_t)*33);
                    if( ed2k_digest(filename, digest) ) {
                        printf("%s", digest);
                    }
                    free(digest);
                    return 1;
                }
            case 'l': {
                    uint8_t *digest = (uint8_t*)malloc(sizeof(uint8_t)*33);
                    memset(digest, 0, sizeof(uint8_t)*33);
                    if( ed2k_digest(filename, digest) ) {
                        uint8_t *link = ed2k_link(filename, digest);
                        printf("%s", link);
                        free(link);
                    }
                    free(digest);
                    return 1;
                }
            case 'h': {
                    usage();
                }
            case '?': {
                    break;
                }
            default: {
                    abort();
                }
        }
    }
    
    return 0;
}

