/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file ed2k.c
* @brief ED2K
* @author AlannY
* @date 2007-10-10
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <dbug.h>

#include "i18n.h"
#include "digest.h"
#include "ff/md4.h"

#include "readhash.h"
#include "file.h"
#include "gui/progress.h"

#include "ff/ed2k.h"

uint8_t ed2k_digest(uint8_t *filename, uint8_t *digest) {
    DBUG_ENTER("ed2k_digest");

    assert(filename);

    ed2k_context context;
    uint8_t buffer[4096];
    uint8_t *res = (uint8_t*)malloc(sizeof(uint8_t)*33);
    memset(res, 0, sizeof(uint8_t)*33);
    uint8_t *t = (uint8_t*)malloc(sizeof(uint8_t)*16);
    memset(t, 0, sizeof(uint8_t)*16);

    set_status((uint8_t*)_("ED2K hash check"));

    uint64_t max = get_filesize(filename);

    file_t *file = file_open(filename, (uint8_t*)"rb");
    if( file ) {
        ed2k_init(&context);
        
        uint64_t k = 0;
        while( 1 ) {
            if( status->flags & STATUS_ABORT ) {
                set_progress(max, max);
                set_status((uint8_t*)_("Done"));
                DBUG_RETURN(FALSE);
            }
            uint64_t i = file_read(file, buffer, 4096);
            if( i > 0 ) {
                ed2k_update(&context, buffer, i);
                k = k + i;
                set_progress(k, max);
            } else {
                break;
            }
        }
        ed2k_final(&context, t);

        file_close(file);
    }

    sprintf((char*)digest, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15]);
    digest[32] = '\0';
    free(t);

    set_status((uint8_t*)_("Done"));

    DBUG_RETURN(TRUE);
}

uint8_t* ed2k_link(uint8_t *filename, uint8_t *ed2k) {
    DBUG_ENTER("ed2k_link");

    assert(filename);
    assert(ed2k);

    uint8_t *f = get_filename(filename);

    uint32_t len = 128;
    len += strlen((char*)f);

    uint32_t size = get_filesize(filename);

    uint8_t *res = (uint8_t*)malloc(sizeof(uint8_t)*len);

    sprintf((char*)res, "ed2k://|file|%s|%u|%s|", f, size, ed2k);

    DBUG_RETURN(res);
}

void ed2k_init(ed2k_context *context) {
    md4_init(&context->context);
    md4_init(&context->inner);
    context->filesize=0;
    context->blockleft=9728000u;
}

void ed2k_update(ed2k_context *context, uint8_t *input, uint32_t length) {
    uint8_t inner[16];
    uint8_t *bufp;
    
    bufp=input;
    while( length+0U > context->blockleft ) {
        md4_update(&context->inner, bufp, context->blockleft);
        bufp += context->blockleft;
        length -= context->blockleft;
        context->blockleft = 9728000u;

        md4_final(&context->inner, inner);
        md4_update(&context->context, inner, 16);
        md4_init(&context->inner);
    }
    if( length > 0 ) {
        md4_update(&context->inner, bufp, length);
        context->blockleft-=length; context->filesize+=length;
    }
}

void ed2k_final(ed2k_context *context, uint8_t *digest) {
    uint8_t inner[16];

    md4_final(&context->inner, inner);

    if( context->filesize > 9728000u ) {
        if (context->blockleft != 9728000u ) {
            md4_update(&context->context, inner, 16);
        }

        md4_final(&context->context, digest);
    } else {
        memcpy(digest, inner, 16);
    }
}

