/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file md4.c
* @brief MD4
* @author AlannY
* @date 2007-10-10
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <dbug.h>

#include "i18n.h"
#include "file.h"
#include "digest.h"

#include "readhash.h"
#include "file.h"
#include "gui/progress.h"

#include "ff/md4.h"

static void md4_transform(uint32_t state[4], uint8_t block[64]);

uint8_t md4_digest(uint8_t *filename, uint8_t *digest) {
    DBUG_ENTER("md4_digest");
    
    /* strlen(digest) = 33*/

    assert(filename);
    
    uint8_t buffer[4096];
    md4_context context;
    uint8_t *t = (uint8_t*)malloc(sizeof(uint8_t)*16);
    memset(t, 0, sizeof(uint8_t)*16);

    set_status((uint8_t*)_("MD4 hash check"));

    uint64_t max = get_filesize(filename);

    file_t *file = file_open(filename, (uint8_t*)"rb");
    if( file ) {
        md4_init(&context);

        uint64_t k = 0;
        while( 1 ) {
            if( status->flags & STATUS_ABORT ) {
                set_progress(max, max);
                set_status((uint8_t*)_("Done"));
                DBUG_RETURN(FALSE);
            }
            uint64_t i = file_read(file, buffer, 4096);
            if( i > 0 ) {
                md4_update(&context, buffer, i);
                k = k + i;
                set_progress(k, max);
            } else {
                break;
            }
        }
        md4_final(&context, t);

        file_close(file);
    }

    sprintf((char*)digest, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15]);
    digest[32] = '\0';
    free(t);

    set_status((uint8_t*)_("Done"));

    DBUG_RETURN(TRUE);
}

#define MD4_S11 3
#define MD4_S12 7
#define MD4_S13 11
#define MD4_S14 19
#define MD4_S21 3
#define MD4_S22 5
#define MD4_S23 9
#define MD4_S24 13
#define MD4_S31 3
#define MD4_S32 9
#define MD4_S33 11
#define MD4_S34 15

static uint8_t MD_PADDING[64] = {
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#define MD4_F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD4_G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define MD4_H(x, y, z) ((x) ^ (y) ^ (z))

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

#define MD4_FF(a, b, c, d, x, s) { (a) += MD4_F ((b), (c), (d)) + (x); (a) = ROTATE_LEFT((a), (s)); }
#define MD4_GG(a, b, c, d, x, s) { (a) += MD4_G ((b), (c), (d)) + (x) + (uint32_t)0x5a827999; (a) = ROTATE_LEFT((a), (s)); }
#define MD4_HH(a, b, c, d, x, s) { (a) += MD4_H ((b), (c), (d)) + (x) + (uint32_t)0x6ed9eba1; (a) = ROTATE_LEFT((a), (s)); }

void md4_init(md4_context *context) {
    context->count[0] = context->count[1] = 0;

    context->state[0] = 0x67452301;
    context->state[1] = 0xefcdab89;
    context->state[2] = 0x98badcfe;
    context->state[3] = 0x10325476;
}

void md4_update(md4_context *context, uint8_t *input, uint32_t inputLen) {
    uint32_t i, index, partLen;

    index = (uint32_t)((context->count[0] >> 3) & 0x3F);
    if( (context->count[0] += ((uint32_t)inputLen << 3) ) < ((uint32_t)inputLen << 3)) {
        context->count[1]++;
    }
    context->count[1] += ((uint32_t)inputLen >> 29);

    partLen = 64 - index;

    if (inputLen >= partLen) {
        memcpy(&context->buffer[index], input, partLen);
        md4_transform(context->state, context->buffer);

        for( i = partLen; i + 63 < inputLen; i += 64 ) {
            md4_transform(context->state, &input[i]);
        }
        
        index = 0;
    } else {
        i = 0;
    }

    memcpy(&context->buffer[index], &input[i], inputLen-i);
}

void md4_final(md4_context *context, uint8_t *digest) {
    uint8_t bits[8];
    uint32_t index, padLen;

    mdX_encode(bits, context->count, 8);

    index = (uint32_t)((context->count[0] >> 3) & 0x3f);
    padLen = (index < 56) ? (56 - index) : (120 - index);
    md4_update(context, MD_PADDING, padLen);
    md4_update(context, bits, 8);

    mdX_encode(digest, context->state, 16);

    memset(context, 0, sizeof (*context));
}

static void md4_transform(uint32_t state[4], uint8_t block[64]) {
    uint32_t a = state[0];
    uint32_t b = state[1];
    uint32_t c = state[2];
    uint32_t d = state[3];
    uint32_t x[16];

    mdX_decode(x, block, 64);

    MD4_FF(a, b, c, d, x[ 0], MD4_S11);
    MD4_FF(d, a, b, c, x[ 1], MD4_S12);
    MD4_FF(c, d, a, b, x[ 2], MD4_S13);
    MD4_FF(b, c, d, a, x[ 3], MD4_S14);
    MD4_FF(a, b, c, d, x[ 4], MD4_S11);
    MD4_FF(d, a, b, c, x[ 5], MD4_S12);
    MD4_FF(c, d, a, b, x[ 6], MD4_S13);
    MD4_FF(b, c, d, a, x[ 7], MD4_S14);
    MD4_FF(a, b, c, d, x[ 8], MD4_S11);
    MD4_FF(d, a, b, c, x[ 9], MD4_S12);
    MD4_FF(c, d, a, b, x[10], MD4_S13);
    MD4_FF(b, c, d, a, x[11], MD4_S14);
    MD4_FF(a, b, c, d, x[12], MD4_S11);
    MD4_FF(d, a, b, c, x[13], MD4_S12);
    MD4_FF(c, d, a, b, x[14], MD4_S13);
    MD4_FF(b, c, d, a, x[15], MD4_S14);

    MD4_GG (a, b, c, d, x[ 0], MD4_S21);
    MD4_GG (d, a, b, c, x[ 4], MD4_S22);
    MD4_GG (c, d, a, b, x[ 8], MD4_S23);
    MD4_GG (b, c, d, a, x[12], MD4_S24);
    MD4_GG (a, b, c, d, x[ 1], MD4_S21);
    MD4_GG (d, a, b, c, x[ 5], MD4_S22);
    MD4_GG (c, d, a, b, x[ 9], MD4_S23);
    MD4_GG (b, c, d, a, x[13], MD4_S24);
    MD4_GG (a, b, c, d, x[ 2], MD4_S21);
    MD4_GG (d, a, b, c, x[ 6], MD4_S22);
    MD4_GG (c, d, a, b, x[10], MD4_S23);
    MD4_GG (b, c, d, a, x[14], MD4_S24);
    MD4_GG (a, b, c, d, x[ 3], MD4_S21);
    MD4_GG (d, a, b, c, x[ 7], MD4_S22);
    MD4_GG (c, d, a, b, x[11], MD4_S23);
    MD4_GG (b, c, d, a, x[15], MD4_S24);

    MD4_HH (a, b, c, d, x[ 0], MD4_S31);
    MD4_HH (d, a, b, c, x[ 8], MD4_S32);
    MD4_HH (c, d, a, b, x[ 4], MD4_S33);
    MD4_HH (b, c, d, a, x[12], MD4_S34);
    MD4_HH (a, b, c, d, x[ 2], MD4_S31);
    MD4_HH (d, a, b, c, x[10], MD4_S32);
    MD4_HH (c, d, a, b, x[ 6], MD4_S33);
    MD4_HH (b, c, d, a, x[14], MD4_S34);
    MD4_HH (a, b, c, d, x[ 1], MD4_S31);
    MD4_HH (d, a, b, c, x[ 9], MD4_S32);
    MD4_HH (c, d, a, b, x[ 5], MD4_S33);
    MD4_HH (b, c, d, a, x[13], MD4_S34);
    MD4_HH (a, b, c, d, x[ 3], MD4_S31);
    MD4_HH (d, a, b, c, x[11], MD4_S32);
    MD4_HH (c, d, a, b, x[ 7], MD4_S33);
    MD4_HH (b, c, d, a, x[15], MD4_S34);

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;

    memset(x, 0, sizeof(x));
}

