/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file sha1.c
* @brief SHA1
* @author AlannY
* @date 2007-10-10
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <dbug.h>

#include "i18n.h"

#include "readhash.h"
#include "file.h"
#include "gui/progress.h"

#include "sha1.h"

static void sha1_process_message_block(sha1_context *context);
static void sha1_pad_message(sha1_context *context);

uint8_t sha1_digest(uint8_t *filename, uint8_t *digest) {
    DBUG_ENTER("sha1_digest");

    assert(filename);

    uint8_t buffer[4096];
    sha1_context context;
    uint8_t *res = (uint8_t*)malloc(sizeof(uint8_t)*41);
    memset(res, 0, sizeof(uint8_t)*41);
    uint8_t *t = (uint8_t*)malloc(sizeof(uint8_t)*20);
    memset(t, 0, sizeof(uint8_t)*20);

    set_status((uint8_t*)_("SHA1 hash check"));

    uint64_t max = get_filesize(filename);

    file_t *file = file_open(filename, (uint8_t*)"rb");
    if( file ) {
        sha1_init(&context);
        
        uint64_t k = 0;
        while( 1 ) {
            if( status->flags & STATUS_ABORT ) {
                set_progress(max, max);
                set_status((uint8_t*)_("Done"));
                DBUG_RETURN(FALSE);
            }
            uint64_t i = file_read(file, buffer, 4096);
            if( i > 0 ) {
                sha1_update(&context, buffer, i);
                k = k + i;
                set_progress(k, max);
            } else {
                break;
            }
        }
        sha1_final(&context, t);

        file_close(file);
    }

    sprintf((char*)digest, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15], t[16], t[17], t[18], t[19]);
    digest[40] = '\0';
    free(t);

    set_status((uint8_t*)_("Done"));

    DBUG_RETURN(TRUE);
}

#define sha1_circular_shift(bits,word) (((word) << (bits)) | ((word) >> (32-(bits))))

void sha1_init(sha1_context *context) {
    context->length_low  = 0;
    context->length_high = 0;
    context->block_index = 0;

    context->hash[0] = 0x67452301;
    context->hash[1] = 0xEFCDAB89;
    context->hash[2] = 0x98BADCFE;
    context->hash[3] = 0x10325476;
    context->hash[4] = 0xC3D2E1F0;

    context->computed  = 0;
    context->corrupted = 0;
}

void sha1_final(sha1_context *context, uint8_t digest[20]) {
    int i;

    if( !context->computed ) {
        sha1_pad_message(context);
        for(i=0; i<64; ++i)
        {
            context->block[i] = 0;
        }
        context->length_low = 0;
        context->length_high = 0;
        context->computed = 1;
    }

    for( i = 0; i < 20; i++ ) {
        digest[i] = context->hash[i>>2] >> 8 * ( 3 - ( i & 0x03 ) );
    }
}

void sha1_update(sha1_context *context, const uint8_t  *message_array, uint32_t length) {
    while(length-- && !context->corrupted) {
        context->block[context->block_index++] = (*message_array & 0xFF);

        context->length_low += 8;
        if( context->length_low == 0 ) {
            context->length_high++;
            if( context->length_high == 0) {
                return;
            }
        }

        if (context->block_index == 64) {
            sha1_process_message_block(context);
        }

        message_array++;
    }
}

static void sha1_process_message_block(sha1_context *context) {
    const uint32_t K[] =    {
                                    0x5A827999,
                                    0x6ED9EBA1,
                                    0x8F1BBCDC,
                                    0xCA62C1D6
                                };
    int i;
    uint32_t temp;
    uint32_t W[80];
    uint32_t A, B, C, D, E;

    for( i = 0; i < 16; i++ ) {
        W[i] = context->block[i * 4] << 24;
        W[i] |= context->block[i * 4 + 1] << 16;
        W[i] |= context->block[i * 4 + 2] << 8;
        W[i] |= context->block[i * 4 + 3];
    }

    for( i = 16; i < 80; i++ ) {
        W[i] = sha1_circular_shift(1,W[i-3] ^ W[i-8] ^ W[i-14] ^ W[i-16]);
    }

    A = context->hash[0];
    B = context->hash[1];
    C = context->hash[2];
    D = context->hash[3];
    E = context->hash[4];

    for( i = 0; i < 20; i++ ) {
        temp =  sha1_circular_shift(5,A) + ((B & C) | ((~B) & D)) + E + W[i] + K[0];
        E = D;
        D = C;
        C = sha1_circular_shift(30,B);
        B = A;
        A = temp;
    }

    for( i = 20; i < 40; i++ ) {
        temp = sha1_circular_shift(5,A) + (B ^ C ^ D) + E + W[i] + K[1];
        E = D;
        D = C;
        C = sha1_circular_shift(30,B);
        B = A;
        A = temp;
    }

    for( i = 40; i < 60; i++ ) {
        temp = sha1_circular_shift(5,A) + ((B & C) | (B & D) | (C & D)) + E + W[i] + K[2];
        E = D;
        D = C;
        C = sha1_circular_shift(30,B);
        B = A;
        A = temp;
    }

    for( i = 60; i < 80; i++ ) {
        temp = sha1_circular_shift(5,A) + (B ^ C ^ D) + E + W[i] + K[3];
        E = D;
        D = C;
        C = sha1_circular_shift(30,B);
        B = A;
        A = temp;
    }

    context->hash[0] += A;
    context->hash[1] += B;
    context->hash[2] += C;
    context->hash[3] += D;
    context->hash[4] += E;

    context->block_index = 0;
}

static void sha1_pad_message(sha1_context *context) {
    if (context->block_index > 55) {
        context->block[context->block_index++] = 0x80;
        while(context->block_index < 64) {
            context->block[context->block_index++] = 0;
        }

        sha1_process_message_block(context);

        while(context->block_index < 56) {
            context->block[context->block_index++] = 0;
        }
    } else {
        context->block[context->block_index++] = 0x80;
        while(context->block_index < 56) {
            context->block[context->block_index++] = 0;
        }
    }

    context->block[56] = context->length_high >> 24;
    context->block[57] = context->length_high >> 16;
    context->block[58] = context->length_high >> 8;
    context->block[59] = context->length_high;
    context->block[60] = context->length_low >> 24;
    context->block[61] = context->length_low >> 16;
    context->block[62] = context->length_low >> 8;
    context->block[63] = context->length_low;

    sha1_process_message_block(context);
}
