/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file sha512.c
* @brief SHA-512
* @author AlannY
* @date 2007-10-20
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <dbug.h>

#include "i18n.h"

#include "readhash.h"
#include "file.h"
#include "gui/progress.h"

#include "ff/sha512.h"

static void sha_transform(sha512_context *context);

uint8_t sha512_digest(uint8_t *filename, uint8_t *digest) {
    DBUG_ENTER("sha512_digest");

    assert(filename);

    uint8_t buffer[4096];
    sha512_context context;
    uint8_t *t = (uint8_t*)malloc(sizeof(uint8_t)*64);
    memset(t, 0, sizeof(uint8_t)*64);

    set_status((uint8_t*)_("SHA512 hash check"));

    uint64_t max = get_filesize(filename);
    
    uint8_t mode[] = "rb";
    file_t *file = file_open(filename, mode);
    if( file ) {
        sha512_init(&context);
        
        uint64_t k = 0;
        while( 1 ) {
            if( status->flags & STATUS_ABORT ) {
                set_progress(max, max);
                set_status((uint8_t*)_("Done"));
                DBUG_RETURN(FALSE);
            }
            uint64_t i = file_read(file, buffer, 4096);
            if( i > 0 ) {
                sha512_update(&context, buffer, i);
                k = k + i;
                set_progress(k, max);
            } else {
                break;
            }
        }
        sha512_final(&context);
        sha512_unpack_digest(&context, t);

        file_close(file);
    }

    sprintf((char*)digest, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15], t[16], t[17], t[18], t[19], t[20], t[21], t[22], t[23], t[24], t[25], t[26], t[27], t[28], t[29], t[30], t[31], t[32], t[33], t[34], t[35], t[36], t[37], t[38], t[39], t[40], t[41], t[42], t[43], t[44], t[45], t[46], t[47], t[48], t[49], t[50], t[51], t[52], t[53], t[54], t[55], t[56], t[57], t[58], t[59], t[60], t[61], t[62], t[63]);
    digest[128] = '\0';
    free(t);

    set_status((uint8_t*)_("Done"));

    DBUG_RETURN(TRUE);
}

uint8_t sha384_digest(uint8_t *filename, uint8_t *digest) {
    DBUG_ENTER("sha384_digest");

    assert(filename);

    uint8_t buffer[4096];
    sha512_context context;
    uint8_t *t = (uint8_t*)malloc(sizeof(uint8_t)*48);
    memset(t, 0, sizeof(uint8_t)*48);

    set_status((uint8_t*)_("SHA384 hash check"));

    uint64_t max = get_filesize(filename);

    file_t *file = file_open(filename, (uint8_t*)"rb");
    if( file ) {
        sha384_init(&context);
        
        uint64_t k = 0;
        while( 1 ) {
            if( status->flags & STATUS_ABORT ) {
                set_progress(max, max);
                set_status((uint8_t*)_("Done"));
                DBUG_RETURN(FALSE);
            }
            uint64_t i = file_read(file, buffer, 4096);
            if( i > 0 ) {
                sha512_update(&context, buffer, i);
                k = k + i;
                set_progress(k, max);
            } else {
                break;
            }
        }
        sha512_final(&context);
        sha384_unpack_digest(&context, t);

        file_close(file);
    }

    sprintf((char*)digest, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X", t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15], t[16], t[17], t[18], t[19], t[20], t[21], t[22], t[23], t[24], t[25], t[26], t[27], t[28], t[29], t[30], t[31], t[32], t[33], t[34], t[35], t[36], t[37], t[38], t[39], t[40], t[41], t[42], t[43], t[44], t[45], t[46], t[47]);
    digest[96] = '\0';
    free(t);

    set_status((uint8_t*)_("Done"));

    DBUG_RETURN(TRUE);
}

#define TRUNC32(x)	((Uint32)((x) & 0xffffffffLL))

/* 64-bit rotate to the RIGHT */
#define ROT64(x,n)	(((x >> n) | (x << (64 - n))))

#define CH(x, y, z) (((x) & (y))^(~(x) & (z)))
#define MAJ(x, y, z)(((x) & (y))^((x) & (z))^((y) & (z)))

/* upper-case sigma functions in SHA spec */
#define USIG0(x) (ROT64(x, 28)^ROT64(x, 34)^ROT64(x, 39))
#define USIG1(x) (ROT64(x, 14)^ROT64(x, 18)^ROT64(x, 41))

/* lower-case sigma functions in SHA spec */
#define LSIG0(x) (ROT64(x, 1)^ROT64(x, 8)^(x >> 7))
#define LSIG1(x) (ROT64(x, 19)^ROT64(x, 61)^(x >> 6))

/* Constants */
static uint64_t K[80] = {
  0x428a2f98d728ae22LL, 0x7137449123ef65cdLL,
  0xb5c0fbcfec4d3b2fLL, 0xe9b5dba58189dbbcLL,
  0x3956c25bf348b538LL, 0x59f111f1b605d019LL,
  0x923f82a4af194f9bLL, 0xab1c5ed5da6d8118LL,
  0xd807aa98a3030242LL, 0x12835b0145706fbeLL,
  0x243185be4ee4b28cLL, 0x550c7dc3d5ffb4e2LL,
  0x72be5d74f27b896fLL, 0x80deb1fe3b1696b1LL,
  0x9bdc06a725c71235LL, 0xc19bf174cf692694LL,
  0xe49b69c19ef14ad2LL, 0xefbe4786384f25e3LL,
  0x0fc19dc68b8cd5b5LL, 0x240ca1cc77ac9c65LL,
  0x2de92c6f592b0275LL, 0x4a7484aa6ea6e483LL,
  0x5cb0a9dcbd41fbd4LL, 0x76f988da831153b5LL,
  0x983e5152ee66dfabLL, 0xa831c66d2db43210LL,
  0xb00327c898fb213fLL, 0xbf597fc7beef0ee4LL,
  0xc6e00bf33da88fc2LL, 0xd5a79147930aa725LL,
  0x06ca6351e003826fLL, 0x142929670a0e6e70LL,
  0x27b70a8546d22ffcLL, 0x2e1b21385c26c926LL,
  0x4d2c6dfc5ac42aedLL, 0x53380d139d95b3dfLL,
  0x650a73548baf63deLL, 0x766a0abb3c77b2a8LL,
  0x81c2c92e47edaee6LL, 0x92722c851482353bLL,
  0xa2bfe8a14cf10364LL, 0xa81a664bbc423001LL,
  0xc24b8b70d0f89791LL, 0xc76c51a30654be30LL,
  0xd192e819d6ef5218LL, 0xd69906245565a910LL,
  0xf40e35855771202aLL, 0x106aa07032bbd1b8LL,
  0x19a4c116b8d2d0c8LL, 0x1e376c085141ab53LL,
  0x2748774cdf8eeb99LL, 0x34b0bcb5e19b48a8LL,
  0x391c0cb3c5c95a63LL, 0x4ed8aa4ae3418acbLL,
  0x5b9cca4f7763e373LL, 0x682e6ff3d6b2b8a3LL,
  0x748f82ee5defb2fcLL, 0x78a5636f43172f60LL,
  0x84c87814a1f0ab72LL, 0x8cc702081a6439ecLL,
  0x90befffa23631e28LL, 0xa4506cebde82bde9LL,
  0xbef9a3f7b2c67915LL, 0xc67178f2e372532bLL,
  0xca273eceea26619cLL, 0xd186b8c721c0c207LL,
  0xeada7dd6cde0eb1eLL, 0xf57d4f7fee6ed178LL,
  0x06f067aa72176fbaLL, 0x0a637dc5a2c898a6LL,
  0x113f9804bef90daeLL, 0x1b710b35131c471bLL,
  0x28db77f523047d84LL, 0x32caab7b40c72493LL,
  0x3c9ebe0a15c9bebcLL, 0x431d67c49c100d4cLL,
  0x4cc5d4becb3e42b6LL, 0x597f299cfc657e2aLL,
  0x5fcb6fab3ad6faecLL, 0x6c44198c4a475817LL
};

static void sha_transform(sha512_context *context) {
    int32_t i, j;
    uint8_t *dp;
    uint64_t T, T1, T2, A, B, C, D, E, F, G, H, W[80];

    dp = context->data;

#ifdef LITTLE_ENDIAN
    for (i = 0; i < 16; ++i) {
        T = *((uint64_t *) dp);
        dp += 8;
        W[i] = ((T << 56) & 0xff00000000000000LL) |
      ((T << 40) & 0x00ff000000000000LL) |
      ((T << 24) & 0x0000ff0000000000LL) |
      ((T <<  8) & 0x000000ff00000000LL) |
      ((T >>  8) & 0x00000000ff000000LL) |
      ((T >> 24) & 0x0000000000ff0000LL) |
      ((T >> 40) & 0x000000000000ff00LL) |
      ((T >> 56) & 0x00000000000000ffLL);
    }
#else
    for (i = 0; i < 16; ++i) {
        T = *((Uint64 *) dp);
        dp += 8;
        W[i] = (T);
  }
#endif

    A = context->digest[0];
    B = context->digest[1];
    C = context->digest[2];
    D = context->digest[3];
    E = context->digest[4];
    F = context->digest[5];
    G = context->digest[6];
    H = context->digest[7];

    for (i=16; i<80; i++)
        W[i] = (LSIG1(W[i-2]) + W[i-7] + LSIG0(W[i-15]) + W[i-16]);

    for (j=0; j<80; j++) {
        T1 = (H + USIG1(E) + CH(E, F, G) + K[j] + W[j]);
        T2 = (USIG0(A) + MAJ(A, B, C));
        H = G;
        G = F;
        F = E;
        E = (D + T1);
        D = C;
        C = B;
        B = A;
        A = (T1 + T2);
    }

    context->digest[0] += A;
    context->digest[1] += B;
    context->digest[2] += C;
    context->digest[3] += D;
    context->digest[4] += E;
    context->digest[5] += F;
    context->digest[6] += G;
    context->digest[7] += H;
}

void sha512_init(sha512_context *context) {
    context->digest[0] = 0x6a09e667f3bcc908LL;
    context->digest[1] = 0xbb67ae8584caa73bLL;
    context->digest[2] = 0x3c6ef372fe94f82bLL;
    context->digest[3] = 0xa54ff53a5f1d36f1LL;
    context->digest[4] = 0x510e527fade682d1LL;
    context->digest[5] = 0x9b05688c2b3e6c1fLL;
    context->digest[6] = 0x1f83d9abfb41bd6bLL;
    context->digest[7] = 0x5be0cd19137e2179LL;
    context->count_lo = 0L;
    context->count_hi = 0L;
    context->local = 0;
    memset((uint8_t*)context->data, 0, 128);
}

void sha384_init(sha512_context *context) {
    context->digest[0] = 0xcbbb9d5dc1059ed8LL;
    context->digest[1] = 0x629a292a367cd507LL;
    context->digest[2] = 0x9159015a3070dd17LL;
    context->digest[3] = 0x152fecd8f70e5939LL;
    context->digest[4] = 0x67332667ffc00b31LL;
    context->digest[5] = 0x8eb44a8768581511LL;
    context->digest[6] = 0xdb0c2e0d64f98fa7LL;
    context->digest[7] = 0x47b5481dbefa4fa4LL;
    context->count_lo = 0L;
    context->count_hi = 0L;
    context->local = 0;
    memset((uint8_t*)context->data, 0, 128);
}

void sha512_update(sha512_context *context, uint8_t *buffer, int count) {
    int32_t i;
    uint64_t clo;

    clo = context->count_lo + (count << 3);
    if (clo < context->count_lo) {
        context->count_hi++;
    }
    context->count_lo = clo;
    if( context->local ) {
        i = 128 - context->local;
        if( i > count ) {
            i = count;
        }
        memcpy(((uint8_t*)context->data) + context->local, buffer, i);
        count -= i;
        buffer += i;
        context->local += i;
        if( context->local == 128 ) {
            sha_transform(context);
        } else {
            return;
        }
    }
    
    while( count >= 128 ) {
        memcpy(context->data, buffer, 128);
        buffer += 128;
        count -= 128;
        sha_transform(context);
    }
    
    memcpy(context->data, buffer, count);
    context->local = count;
}

void sha512_final(sha512_context *context) {
    int32_t count;
    uint64_t lo_bit_count, hi_bit_count;

    lo_bit_count = context->count_lo;
    hi_bit_count = context->count_hi;
    count = (int32_t)((lo_bit_count >> 3) & 0x7f);
    ((uint8_t*)context->data)[count++] = 0x80;
    if( count > 128 - 16) {
        memset(((uint8_t*)context->data)+count, 0, 128 - count);
        sha_transform(context);
        memset((uint8_t*)context->data, 0, 128 - 16);
    } else {
        memset(((uint8_t*)context->data) + count, 0, 128 - count - 16);
    }

    context->data[112] = (hi_bit_count >> 56) & 0xff;
    context->data[113] = (hi_bit_count >> 48) & 0xff;
    context->data[114] = (hi_bit_count >> 40) & 0xff;
    context->data[115] = (hi_bit_count >> 32) & 0xff;
    context->data[116] = (hi_bit_count >> 24) & 0xff;
    context->data[117] = (hi_bit_count >> 16) & 0xff;
    context->data[118] = (hi_bit_count >>  8) & 0xff;
    context->data[119] = (hi_bit_count >>  0) & 0xff;
    context->data[120] = (lo_bit_count >> 56) & 0xff;
    context->data[121] = (lo_bit_count >> 48) & 0xff;
    context->data[122] = (lo_bit_count >> 40) & 0xff;
    context->data[123] = (lo_bit_count >> 32) & 0xff;
    context->data[124] = (lo_bit_count >> 24) & 0xff;
    context->data[125] = (lo_bit_count >> 16) & 0xff;
    context->data[126] = (lo_bit_count >>  8) & 0xff;
    context->data[127] = (lo_bit_count >>  0) & 0xff;

    sha_transform(context);
}

void sha384_unpack_digest(sha512_context *context, uint8_t *digest) {
    digest[ 0] = (uint8_t)((context->digest[0] >> 56) & 0xff);
    digest[ 1] = (uint8_t)((context->digest[0] >> 48) & 0xff);
    digest[ 2] = (uint8_t)((context->digest[0] >> 40) & 0xff);
    digest[ 3] = (uint8_t)((context->digest[0] >> 32) & 0xff);
    digest[ 4] = (uint8_t)((context->digest[0] >> 24) & 0xff);
    digest[ 5] = (uint8_t)((context->digest[0] >> 16) & 0xff);
    digest[ 6] = (uint8_t)((context->digest[0] >>  8) & 0xff);
    digest[ 7] = (uint8_t)((context->digest[0]      ) & 0xff);
    digest[ 8] = (uint8_t)((context->digest[1] >> 56) & 0xff);
    digest[ 9] = (uint8_t)((context->digest[1] >> 48) & 0xff);
    digest[10] = (uint8_t)((context->digest[1] >> 40) & 0xff);
    digest[11] = (uint8_t)((context->digest[1] >> 32) & 0xff);
    digest[12] = (uint8_t)((context->digest[1] >> 24) & 0xff);
    digest[13] = (uint8_t)((context->digest[1] >> 16) & 0xff);
    digest[14] = (uint8_t)((context->digest[1] >>  8) & 0xff);
    digest[15] = (uint8_t)((context->digest[1]      ) & 0xff);
    digest[16] = (uint8_t)((context->digest[2] >> 56) & 0xff);
    digest[17] = (uint8_t)((context->digest[2] >> 48) & 0xff);
    digest[18] = (uint8_t)((context->digest[2] >> 40) & 0xff);
    digest[19] = (uint8_t)((context->digest[2] >> 32) & 0xff);
    digest[20] = (uint8_t)((context->digest[2] >> 24) & 0xff);
    digest[21] = (uint8_t)((context->digest[2] >> 16) & 0xff);
    digest[22] = (uint8_t)((context->digest[2] >>  8) & 0xff);
    digest[23] = (uint8_t)((context->digest[2]      ) & 0xff);
    digest[24] = (uint8_t)((context->digest[3] >> 56) & 0xff);
    digest[25] = (uint8_t)((context->digest[3] >> 48) & 0xff);
    digest[26] = (uint8_t)((context->digest[3] >> 40) & 0xff);
    digest[27] = (uint8_t)((context->digest[3] >> 32) & 0xff);
    digest[28] = (uint8_t)((context->digest[3] >> 24) & 0xff);
    digest[29] = (uint8_t)((context->digest[3] >> 16) & 0xff);
    digest[30] = (uint8_t)((context->digest[3] >>  8) & 0xff);
    digest[31] = (uint8_t)((context->digest[3]      ) & 0xff);
    digest[32] = (uint8_t)((context->digest[4] >> 56) & 0xff);
    digest[33] = (uint8_t)((context->digest[4] >> 48) & 0xff);
    digest[34] = (uint8_t)((context->digest[4] >> 40) & 0xff);
    digest[35] = (uint8_t)((context->digest[4] >> 32) & 0xff);
    digest[36] = (uint8_t)((context->digest[4] >> 24) & 0xff);
    digest[37] = (uint8_t)((context->digest[4] >> 16) & 0xff);
    digest[38] = (uint8_t)((context->digest[4] >>  8) & 0xff);
    digest[39] = (uint8_t)((context->digest[4]      ) & 0xff);
    digest[40] = (uint8_t)((context->digest[5] >> 56) & 0xff);
    digest[41] = (uint8_t)((context->digest[5] >> 48) & 0xff);
    digest[42] = (uint8_t)((context->digest[5] >> 40) & 0xff);
    digest[43] = (uint8_t)((context->digest[5] >> 32) & 0xff);
    digest[44] = (uint8_t)((context->digest[5] >> 24) & 0xff);
    digest[45] = (uint8_t)((context->digest[5] >> 16) & 0xff);
    digest[46] = (uint8_t)((context->digest[5] >>  8) & 0xff);
    digest[47] = (uint8_t)((context->digest[5]      ) & 0xff);
}

void sha512_unpack_digest(sha512_context *context, uint8_t *digest) {
    digest[ 0] = (uint8_t) ((context->digest[0] >> 56) & 0xff);
    digest[ 1] = (uint8_t) ((context->digest[0] >> 48) & 0xff);
    digest[ 2] = (uint8_t) ((context->digest[0] >> 40) & 0xff);
    digest[ 3] = (uint8_t) ((context->digest[0] >> 32) & 0xff);
    digest[ 4] = (uint8_t) ((context->digest[0] >> 24) & 0xff);
    digest[ 5] = (uint8_t) ((context->digest[0] >> 16) & 0xff);
    digest[ 6] = (uint8_t) ((context->digest[0] >>  8) & 0xff);
    digest[ 7] = (uint8_t) ((context->digest[0]      ) & 0xff);

    digest[ 8] = (uint8_t) ((context->digest[1] >> 56) & 0xff);
    digest[ 9] = (uint8_t) ((context->digest[1] >> 48) & 0xff);
    digest[10] = (uint8_t) ((context->digest[1] >> 40) & 0xff);
    digest[11] = (uint8_t) ((context->digest[1] >> 32) & 0xff);
    digest[12] = (uint8_t) ((context->digest[1] >> 24) & 0xff);
    digest[13] = (uint8_t) ((context->digest[1] >> 16) & 0xff);
    digest[14] = (uint8_t) ((context->digest[1] >>  8) & 0xff);
    digest[15] = (uint8_t) ((context->digest[1]      ) & 0xff);

    digest[16] = (uint8_t) ((context->digest[2] >> 56) & 0xff);
    digest[17] = (uint8_t) ((context->digest[2] >> 48) & 0xff);
    digest[18] = (uint8_t) ((context->digest[2] >> 40) & 0xff);
    digest[19] = (uint8_t) ((context->digest[2] >> 32) & 0xff);
    digest[20] = (uint8_t) ((context->digest[2] >> 24) & 0xff);
    digest[21] = (uint8_t) ((context->digest[2] >> 16) & 0xff);
    digest[22] = (uint8_t) ((context->digest[2] >>  8) & 0xff);
    digest[23] = (uint8_t) ((context->digest[2]      ) & 0xff);

    digest[24] = (uint8_t) ((context->digest[3] >> 56) & 0xff);
    digest[25] = (uint8_t) ((context->digest[3] >> 48) & 0xff);
    digest[26] = (uint8_t) ((context->digest[3] >> 40) & 0xff);
    digest[27] = (uint8_t) ((context->digest[3] >> 32) & 0xff);
    digest[28] = (uint8_t) ((context->digest[3] >> 24) & 0xff);
    digest[29] = (uint8_t) ((context->digest[3] >> 16) & 0xff);
    digest[30] = (uint8_t) ((context->digest[3] >>  8) & 0xff);
    digest[31] = (uint8_t) ((context->digest[3]      ) & 0xff);

    digest[32] = (uint8_t) ((context->digest[4] >> 56) & 0xff);
    digest[33] = (uint8_t) ((context->digest[4] >> 48) & 0xff);
    digest[34] = (uint8_t) ((context->digest[4] >> 40) & 0xff);
    digest[35] = (uint8_t) ((context->digest[4] >> 32) & 0xff);
    digest[36] = (uint8_t) ((context->digest[4] >> 24) & 0xff);
    digest[37] = (uint8_t) ((context->digest[4] >> 16) & 0xff);
    digest[38] = (uint8_t) ((context->digest[4] >>  8) & 0xff);
    digest[39] = (uint8_t) ((context->digest[4]      ) & 0xff);

    digest[40] = (uint8_t) ((context->digest[5] >> 56) & 0xff);
    digest[41] = (uint8_t) ((context->digest[5] >> 48) & 0xff);
    digest[42] = (uint8_t) ((context->digest[5] >> 40) & 0xff);
    digest[43] = (uint8_t) ((context->digest[5] >> 32) & 0xff);
    digest[44] = (uint8_t) ((context->digest[5] >> 24) & 0xff);
    digest[45] = (uint8_t) ((context->digest[5] >> 16) & 0xff);
    digest[46] = (uint8_t) ((context->digest[5] >>  8) & 0xff);
    digest[47] = (uint8_t) ((context->digest[5]      ) & 0xff);

    digest[48] = (uint8_t) ((context->digest[6] >> 56) & 0xff);
    digest[49] = (uint8_t) ((context->digest[6] >> 48) & 0xff);
    digest[50] = (uint8_t) ((context->digest[6] >> 40) & 0xff);
    digest[51] = (uint8_t) ((context->digest[6] >> 32) & 0xff);
    digest[52] = (uint8_t) ((context->digest[6] >> 24) & 0xff);
    digest[53] = (uint8_t) ((context->digest[6] >> 16) & 0xff);
    digest[54] = (uint8_t) ((context->digest[6] >>  8) & 0xff);
    digest[55] = (uint8_t) ((context->digest[6]      ) & 0xff);

    digest[56] = (uint8_t) ((context->digest[7] >> 56) & 0xff);
    digest[57] = (uint8_t) ((context->digest[7] >> 48) & 0xff);
    digest[58] = (uint8_t) ((context->digest[7] >> 40) & 0xff);
    digest[59] = (uint8_t) ((context->digest[7] >> 32) & 0xff);
    digest[60] = (uint8_t) ((context->digest[7] >> 24) & 0xff);
    digest[61] = (uint8_t) ((context->digest[7] >> 16) & 0xff);
    digest[62] = (uint8_t) ((context->digest[7] >>  8) & 0xff);
    digest[63] = (uint8_t) ((context->digest[7]      ) & 0xff);
}

