/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file file.c
* @brief File
* @author AlannY
* @date 2007-10-14
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h> 
#include <sys/stat.h>
#include <assert.h>

#include <dbug.h>

#ifdef USE_GNOMEVFS
#include <libgnomevfs/gnome-vfs.h>
#endif

#include "file.h"

inline uint8_t* read_file(uint8_t *filename) {
    DBUG_ENTER("read_file");

    DBUG_ASSERT(filename);

    uint8_t *buf = NULL;
    uint64_t length;

    file_t *file;
    file = file_open(filename, (uint8_t*)"rb");
    if( file ) {
        file_seek(file, 0, SEEK_END);
        length = file_tell(file);
        file_seek(file, 0, SEEK_SET);
        buf = (uint8_t*)malloc(sizeof(uint8_t)*length);
        (void)file_read(file, buf, length);
        file_close(file);
    }

    DBUG_RETURN(buf);
}

inline file_t* file_open(uint8_t *filename, uint8_t *mode) {
    DBUG_ENTER("file_open");

    DBUG_ASSERT(filename);
    DBUG_ASSERT(mode);

    file_t *file;
#ifdef USE_GNOMEVFS
    if( (mode[0] == 'r') || (mode[1] == 'r') ) {
        (void)gnome_vfs_open(&file, (gchar*)filename, GNOME_VFS_OPEN_READ);
    }
    if( (mode[0] == 'w') || (mode[1] == 'w') ) {
        (void)gnome_vfs_open(&file, (gchar*)filename, GNOME_VFS_OPEN_WRITE);
    }
#else
    file = fopen((char*)filename, (char*)mode);
#endif
    DBUG_RETURN(file);
}

inline uint64_t file_read(file_t *file, uint8_t *data, uint64_t len) {

    DBUG_ASSERT(file);
    DBUG_ASSERT(data);
#ifdef USE_GNOMEVFS
    uint64_t size = 0;
    (void)gnome_vfs_read(file, data, len, &size);
    return size;
#else
    return fread(data, 1, len, file);
#endif
}

inline void file_seek(file_t *file, uint64_t offset, uint8_t position) {
    DBUG_ENTER("file_seek");

    DBUG_ASSERT(file);
#ifdef USE_GNOMEVFS
    GnomeVFSSeekPosition pos = GNOME_VFS_SEEK_START;
    if( position == SEEK_SET ) {
        pos = GNOME_VFS_SEEK_START;
    }
    if( position == SEEK_CUR ) {
        pos = GNOME_VFS_SEEK_CURRENT; 
    }
    if( position == SEEK_END ) {
        pos = GNOME_VFS_SEEK_END;
    }
    (void)gnome_vfs_seek(file, pos, offset);
#else
    (void)fseek(file, offset, position);
#endif
    DBUG_VOID_RETURN;
}

inline uint64_t file_tell(file_t *file) {
    DBUG_ENTER("file_tell");

    DBUG_ASSERT(file);
#ifdef USE_GNOMEVFS
    uint64_t pos = 0;
    (void)gnome_vfs_tell(file, &pos);
    DBUG_RETURN(pos);
#else
    DBUG_RETURN(ftell(file));
#endif
}

inline void file_write(file_t *file, uint8_t *data, uint64_t len) {
    DBUG_ENTER("file_write");

    DBUG_ASSERT(file);
    DBUG_ASSERT(data);
#ifdef USE_GNOMEVFS
    (void)gnome_vfs_write(file, data, len, NULL);
#else
    (void)fwrite(data, 1, len, file);
#endif
    DBUG_VOID_RETURN;
}

inline void file_write_line(file_t *file, uint8_t *data) {
    DBUG_ENTER("file_write_line");

    DBUG_ASSERT(file);
    DBUG_ASSERT(data);

    DBUG_RETURN(file_write(file, data, strlen((char*)data)));
}

inline void file_close(file_t *file) {
    DBUG_ENTER("file_close");

    DBUG_ASSERT(file);
#ifdef USE_GNOMEVFS
    (void)gnome_vfs_close(file);
#else
    fclose(file);
#endif
    DBUG_VOID_RETURN;
}

inline uint8_t* get_filename(uint8_t *filename) {
    DBUG_ENTER("get_filename");

    DBUG_ASSERT(filename);
    uint8_t *f = filename;
    uint32_t i = 0;
    uint8_t has_slash = 0;
    for( ; *f != '\0'; f=f+1 );
    for( ; f != filename; f=f-1 ) {
        if( *f == '\\' ) {
            has_slash = 1;
            break;
        }
        if( *f == '/' ) {
            has_slash = 1;
            break;
        }
        i++;
    }

    if( has_slash )
        f++;

    uint8_t *fn = (uint8_t*)malloc(sizeof(uint8_t)*i);
    strncpy((char*)fn, (char*)f, i);
    *(fn+i) = '\0';

    DBUG_RETURN(fn);
}

inline uint8_t* get_path(uint8_t *filename) {
    DBUG_ENTER("get_path");

    DBUG_ASSERT(filename);
    uint8_t *f = filename;
    uint32_t i = 0;
    for( ; *f != '\0'; f=f+1 );
    for( ; f != filename; f=f-1 ) {
        if( *f == '\\' )
            break;
        if( *f == '/' )
            break;
    }

    uint8_t *f2 = filename;
    for( ; f2 != f; f2=f2+1 ) {
        i += 1;
    }

    uint8_t *fn = (uint8_t*)malloc(sizeof(uint8_t)*(i+1));
    strncpy((char*)fn, (char*)filename, i);
    *(fn+i) = '\0';

    DBUG_RETURN(fn);
}

inline uint8_t* get_ext(uint8_t *filename) {
    DBUG_ENTER("get_ext");

    DBUG_ASSERT(filename);

    uint8_t *f = filename;
    uint32_t i = 0;
    uint8_t dotfound = 0;
    for( ; *f != '\0'; f=f+1 );
    for( ; f != filename; f=f-1 ) {
        if( *f == '.' ) {
            dotfound = 1;
            break;
        }
        i++;
    }
    
    if( dotfound ) {
        f++;

        uint8_t *fn = (uint8_t*)malloc(sizeof(uint8_t)*i);
        strncpy((char*)fn, (char*)f, i);
        *(fn+i) = '\0';
        
        DBUG_RETURN(fn);
    } else {
        DBUG_RETURN(NULL);
    }
}

inline uint64_t get_filesize(uint8_t *filename) {
    DBUG_ENTER("get_filesize");

    DBUG_ASSERT(filename);
#ifdef USE_GNOMEVFS
    GnomeVFSFileInfo file_info;
    (void)gnome_vfs_get_file_info((gchar*)filename, &file_info, GNOME_VFS_FILE_INFO_DEFAULT);
    DBUG_RETURN(file_info.size);
#else
    struct stat file_stat;
    int err = stat((char*)filename, &file_stat);
    if( err != 0)
        DBUG_RETURN(0); 
    DBUG_RETURN(file_stat.st_size);
#endif
}

inline void file_rename(uint8_t *from, uint8_t *to) {
    DBUG_ENTER("file_rename");
    
#ifdef USE_GNOMEVFS
    GnomeVFSURI *from_uri = gnome_vfs_uri_new((gchar*)from);
    GnomeVFSURI *to_uri = gnome_vfs_uri_new((gchar*)to);
    GnomeVFSResult result = gnome_vfs_xfer_uri(from_uri, to_uri, GNOME_VFS_XFER_DEFAULT, GNOME_VFS_XFER_ERROR_MODE_ABORT, GNOME_VFS_XFER_OVERWRITE_MODE_ABORT, NULL, NULL);
    if( result == GNOME_VFS_OK ) {
        GList *list = NULL;
        list = g_list_append(list, from_uri);
        gnome_vfs_xfer_delete_list(list, GNOME_VFS_XFER_ERROR_MODE_ABORT, GNOME_VFS_XFER_DEFAULT, NULL, NULL);
    }
#else
    rename((char*)from, (char*)to);
#endif

    DBUG_VOID_RETURN;
}

