/*
 * Copyright (C) 2007 - AlannY
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/** 
* @file window.c
* @brief Main Window
* @author AlannY
* @date 2007-10-09
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <dbug.h>

#include <gtk/gtk.h>

#include "i18n.h"
#include "file.h"

#include "entry.h"
#include "queue.h"

#include "readhash.h"

#include "gui/progress.h"

#include "gui/window.h"

inline static void _create_window();
static void selection_text(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static void on_checkbox_toggled(GtkWidget *widget, gpointer data);
static void destroy(GtkWidget *widget, gpointer data);
static void selection_delete(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList **rowref_list);
static void on_popup_menu_delete(GtkWidget *widget, gpointer data);
static void show_popup_menu(GtkWidget *treeview, GdkEventButton *event, gpointer data);
static gboolean on_button_pressed(GtkWidget *treeview, GdkEventButton *event, gpointer data);
static gboolean on_popup_menu(GtkWidget *treeview, gpointer userdata);
static void selection_changed(GtkTreeSelection *selection, gpointer data);
static void on_add_file(GtkWidget *widget, gpointer data);
static void on_save_sfv(GtkWidget *widget, gpointer data);
static void selection_put(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static void on_put_crc32(GtkWidget *widget, gpointer data);
static void on_abort(GtkWidget *widget, gpointer data);
static void selection_func(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static void on_crc32(GtkWidget *widget, gpointer data);
static void on_md4(GtkWidget *widget, gpointer data);
static void on_md5(GtkWidget *widget, gpointer data);
static void on_sha1(GtkWidget *widget, gpointer data);
static void on_sha256(GtkWidget *widget, gpointer data);
static void on_sha384(GtkWidget *widget, gpointer data);
static void on_sha512(GtkWidget *widget, gpointer data);
static void on_ed2k(GtkWidget *widget, gpointer data);
static void on_ed2klink(GtkWidget *widget, gpointer data);
static void on_tiger(GtkWidget *widget, gpointer data);
static void on_save_sfv(GtkWidget *widget, gpointer data);

#define PATH_DELIM "/"

struct window_s window;

int init_gui(int *argc, char **argv[]) {
    DBUG_ENTER("init_gui");

    assert(argc);
    assert(argv);
    
#ifndef USE_GNOMEVFS
    g_thread_init(NULL);
#endif

    gdk_threads_init();

    gdk_threads_enter();
    gtk_init(argc, argv);

    _create_window();

    gtk_main();
    gdk_threads_leave();

    DBUG_RETURN(0);
}

static inline void _create_window() {
    DBUG_ENTER("_create_window");
    
    GtkWidget *button;
    GtkWidget *image;
    GtkWidget *vbox1;
    GtkWidget *vbox2;
    GtkWidget *hbox1;
    GtkWidget *hbox2;
    GtkWidget *hbox3;
    GtkWidget *paned;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkWidget *view;
    GtkWidget *view_sw;

    vbox1 = gtk_vbox_new(FALSE, 0);
    vbox2 = gtk_vbox_new(FALSE, 0);
    hbox1 = gtk_hbox_new(FALSE, 0);
    hbox2 = gtk_hbox_new(FALSE, 0);
    hbox3 = gtk_hbox_new(FALSE, 0);

    image = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_LARGE_TOOLBAR);
    button = gtk_button_new_with_mnemonic(_("_Open file"));
    gtk_button_set_image(GTK_BUTTON(button), image);
    g_signal_connect(button, "clicked", G_CALLBACK(on_add_file), NULL);
    gtk_box_pack_start(GTK_BOX(hbox1), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    image = gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_LARGE_TOOLBAR);
    button = gtk_button_new_with_mnemonic(_("_Save SFV"));
    gtk_button_set_image(GTK_BUTTON(button), image);
    g_signal_connect(button, "clicked", G_CALLBACK(on_save_sfv), NULL);
    gtk_box_pack_start(GTK_BOX(hbox1), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    image = gtk_image_new_from_stock(GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_LARGE_TOOLBAR);
    button = gtk_button_new_with_mnemonic(_("_Put CRC32 in filename"));
    gtk_button_set_image(GTK_BUTTON(button), image);
    g_signal_connect(button, "clicked", G_CALLBACK(on_put_crc32), NULL);
    gtk_box_pack_start(GTK_BOX(hbox1), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    image = gtk_image_new_from_stock(GTK_STOCK_CANCEL, GTK_ICON_SIZE_LARGE_TOOLBAR);
    window.abort_button = gtk_button_new_with_mnemonic(_("_Abort"));
    gtk_button_set_image(GTK_BUTTON(window.abort_button), image);
    gtk_widget_set_state(window.abort_button, GTK_STATE_INSENSITIVE);
    g_signal_connect(window.abort_button, "clicked", G_CALLBACK(on_abort), NULL);
    gtk_box_pack_start(GTK_BOX(hbox1), window.abort_button, FALSE, FALSE, 0);
    gtk_widget_show(window.abort_button);

    gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);
    gtk_widget_show(hbox1);

    view_sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(view_sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_set_size_request(view_sw, 600, 300);
    
    window.file_store = gtk_list_store_new(N_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);

    view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(window.file_store));
    g_object_unref(window.file_store);
    g_signal_connect(view, "button-press-event", G_CALLBACK(on_button_pressed), NULL);
    g_signal_connect(view, "popup-menu", G_CALLBACK(on_popup_menu), NULL);

    renderer = gtk_cell_renderer_pixbuf_new();
    column = gtk_tree_view_column_new_with_attributes("", renderer, "pixbuf", PIXBUF_COLUMN, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes(_("File"), renderer, "text", FILE_COLUMN, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes(_("CRC32"), renderer, "text", CRC32_COLUMN, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes(_("Status"), renderer, "text", STATUS_COLUMN, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

    window.select = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
    gtk_tree_selection_set_mode(window.select, GTK_SELECTION_MULTIPLE);
    g_signal_connect(G_OBJECT(window.select), "changed", G_CALLBACK(selection_changed), NULL);

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(view_sw), view);
    gtk_widget_show(view);

    gtk_box_pack_start(GTK_BOX(vbox1), view_sw, TRUE, TRUE, 0);
    gtk_widget_show(view_sw);

    button = gtk_check_button_new_with_label(_("Automaticaly compute CRC32"));
    g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(on_checkbox_toggled), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
    gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    
    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
    gtk_widget_show(hbox2);

    window.crc32_button = gtk_button_new_with_label(_("CRC32"));
    g_signal_connect(G_OBJECT(window.crc32_button), "clicked", G_CALLBACK(on_crc32), NULL);
    gtk_widget_set_state(window.crc32_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.crc32_button, FALSE, FALSE, 0);
    gtk_widget_show(window.crc32_button);

    window.md4_button = gtk_button_new_with_label(_("MD4"));
    g_signal_connect(G_OBJECT(window.md4_button), "clicked", G_CALLBACK(on_md4), NULL);
    gtk_widget_set_state(window.md4_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.md4_button, FALSE, FALSE, 0);
    gtk_widget_show(window.md4_button);
    
    window.md5_button = gtk_button_new_with_label(_("MD5"));
    g_signal_connect(G_OBJECT(window.md5_button), "clicked", G_CALLBACK(on_md5), NULL);
    gtk_widget_set_state(window.md5_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.md5_button, FALSE, FALSE, 0);
    gtk_widget_show(window.md5_button);
    
    window.sha1_button = gtk_button_new_with_label(_("SHA1"));
    g_signal_connect(G_OBJECT(window.sha1_button), "clicked", G_CALLBACK(on_sha1), NULL);
    gtk_widget_set_state(window.sha1_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.sha1_button, FALSE, FALSE, 0);
    gtk_widget_show(window.sha1_button);

    window.sha256_button = gtk_button_new_with_label(_("SHA256"));
    g_signal_connect(G_OBJECT(window.sha256_button), "clicked", G_CALLBACK(on_sha256), NULL);
    gtk_widget_set_state(window.sha256_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.sha256_button, FALSE, FALSE, 0);
    gtk_widget_show(window.sha256_button);

    window.sha384_button = gtk_button_new_with_label(_("SHA384"));
    g_signal_connect(G_OBJECT(window.sha384_button), "clicked", G_CALLBACK(on_sha384), NULL);
    gtk_widget_set_state(window.sha384_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.sha384_button, FALSE, FALSE, 0);
    gtk_widget_show(window.sha384_button);

    window.sha512_button = gtk_button_new_with_label(_("SHA512"));
    g_signal_connect(G_OBJECT(window.sha512_button), "clicked", G_CALLBACK(on_sha512), NULL);
    gtk_widget_set_state(window.sha512_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.sha512_button, FALSE, FALSE, 0);
    gtk_widget_show(window.sha512_button);

    window.ed2k_button = gtk_button_new_with_label(_("ED2K"));
    g_signal_connect(G_OBJECT(window.ed2k_button), "clicked", G_CALLBACK(on_ed2k), NULL);
    gtk_widget_set_state(window.ed2k_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.ed2k_button, FALSE, FALSE, 0);
    gtk_widget_show(window.ed2k_button);

    window.ed2klink_button = gtk_button_new_with_label(_("ED2K Link"));
    g_signal_connect(G_OBJECT(window.ed2klink_button), "clicked", G_CALLBACK(on_ed2klink), NULL);
    gtk_widget_set_state(window.ed2klink_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.ed2klink_button, FALSE, FALSE, 0);
    gtk_widget_show(window.ed2klink_button);

    window.tiger_button = gtk_button_new_with_label(_("Tiger"));
    g_signal_connect(G_OBJECT(window.tiger_button), "clicked", G_CALLBACK(on_tiger), NULL);
    gtk_widget_set_state(window.tiger_button, GTK_STATE_INSENSITIVE);
    gtk_box_pack_start(GTK_BOX(hbox3), window.tiger_button, FALSE, FALSE, 0);
    gtk_widget_show(window.tiger_button);

    gtk_box_pack_start(GTK_BOX(vbox2), hbox3, FALSE, FALSE, 0);
    gtk_widget_show(hbox3);

    window.progressbar = gtk_progress_bar_new();
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(window.progressbar), 1);

    gtk_box_pack_start(GTK_BOX(vbox2), window.progressbar, FALSE, FALSE, 0);
    gtk_widget_show(window.progressbar);
    
    view = gtk_text_view_new();
    window.buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
    window.bold = gtk_text_buffer_create_tag(window.buffer, "bold", "weight", PANGO_WEIGHT_BOLD, NULL);
    window.blue = gtk_text_buffer_create_tag(window.buffer, "blue", "foreground", "blue", NULL);

    view_sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(view_sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_set_size_request(view_sw, 600, 150);

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(view_sw), view);
    gtk_widget_show(view);

    gtk_box_pack_start(GTK_BOX(vbox2), view_sw, TRUE, TRUE, 0);
    gtk_widget_show(view_sw);

    window.statusbar = gtk_statusbar_new();

    gtk_box_pack_start(GTK_BOX(vbox2), window.statusbar, FALSE, FALSE, 0);
    gtk_statusbar_push(GTK_STATUSBAR(window.statusbar), 0, _("Done"));
    gtk_widget_show(window.statusbar);

    paned = gtk_vpaned_new();

    gtk_paned_pack1(GTK_PANED(paned), vbox1, TRUE, TRUE);
    gtk_widget_show(vbox1);

    gtk_paned_pack2(GTK_PANED(paned), vbox2, FALSE, FALSE);
    gtk_widget_show(vbox2);

    window.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window.window), _("readhash"));
    g_signal_connect(G_OBJECT(window.window), "destroy", G_CALLBACK(destroy), NULL);
    gtk_container_add(GTK_CONTAINER(window.window), paned);
    GdkPixbuf *icon = gdk_pixbuf_new_from_file(ICONSDIR"/readhash.svg", NULL);
    gtk_window_set_icon(GTK_WINDOW(window.window), icon);
    gtk_widget_show(paned);
    gtk_widget_show(window.window);

    window.grey = gdk_pixbuf_new_from_file(ICONSDIR"/grey.svg", NULL);
    window.green = gdk_pixbuf_new_from_file(ICONSDIR"/green.svg", NULL);
    window.red = gdk_pixbuf_new_from_file(ICONSDIR"/red.svg", NULL);

    status->flags |= STATUS_GUI;
    
    DBUG_VOID_RETURN;
}

static void on_checkbox_toggled(GtkWidget *widget, gpointer data) {
    status->flags ^= STATUS_AUTO_CRC32;
}

static void destroy(GtkWidget *widget, gpointer data) {
    GtkTreeIter iter;
    gboolean valid;

    status->flags |= STATUS_QUIT;

    valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(window.file_store), &iter);

    while( valid ) {
        entry_t *entry = NULL;
        gtk_tree_model_get(GTK_TREE_MODEL(window.file_store), &iter, ENTRY_COLUMN, &entry, -1);

        entry_free(entry);

        valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(window.file_store), &iter);
    }

    gtk_main_quit();
}

static void on_abort(GtkWidget *widget, gpointer data) {
    status->flags |= STATUS_ABORT;
    queue_flush(queue);
    gtk_widget_set_sensitive(window.abort_button, FALSE);
}

static void selection_delete(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList **rowref_list) {
    GtkTreeRowReference *rowref;
    rowref = gtk_tree_row_reference_new(model, path);
    *rowref_list = g_list_append(*rowref_list, rowref);
}

static void on_popup_menu_delete(GtkWidget *widget, gpointer data) {
    GList *rr_list = NULL;
    GList *node;

    gtk_tree_selection_selected_foreach(window.select, (GtkTreeSelectionForeachFunc)selection_delete, &rr_list);

    for( node = rr_list; node != NULL; node = node->next ) {
        GtkTreePath *path;
        path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data);
        if( path ) {
            GtkTreeIter iter;
            if( gtk_tree_model_get_iter(GTK_TREE_MODEL(window.file_store), &iter, path)) {
                entry_t *entry = NULL;
                gtk_tree_model_get(GTK_TREE_MODEL(window.file_store), &iter, ENTRY_COLUMN, &entry, -1);

                if( entry == cur_entry ) {
                    status->flags |= STATUS_ABORT;
                }

                queue_remove(queue, entry);
                entry_free(entry);

                gtk_list_store_remove(window.file_store, &iter);
            }
            gtk_tree_path_free(path);
        }
    }
    
    g_list_foreach(rr_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free(rr_list);
}

static void show_popup_menu(GtkWidget *treeview, GdkEventButton *event, gpointer data) {
    GtkWidget *menu;
    GtkWidget *menuitem;
    GtkWidget *image;
    
    uint32_t rows = gtk_tree_selection_count_selected_rows(window.select);

    menu = gtk_menu_new();

    image = gtk_image_new_from_stock(GTK_STOCK_DELETE, GTK_ICON_SIZE_MENU);
    menuitem = gtk_image_menu_item_new_with_mnemonic(_("_Delete"));
    if( rows == 0 ) {
        gtk_widget_set_state(menuitem, GTK_STATE_INSENSITIVE);
    }
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), image);
    g_signal_connect(menuitem, "activate", G_CALLBACK(on_popup_menu_delete), treeview);

    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
    gtk_widget_show(menuitem);

    gtk_widget_show(menu);
    
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, (event != NULL) ? event->button : 0, gdk_event_get_time((GdkEvent*)event));
}

static gboolean on_button_pressed(GtkWidget *treeview, GdkEventButton *event, gpointer data) {
    if( (event->type == GDK_BUTTON_PRESS) && (event->button == 3) ) {
        if( gtk_tree_selection_count_selected_rows(window.select) <= 1 ) {
            GtkTreePath *path;
            if( gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(treeview), (gint)event->x, (gint)event->y, &path, NULL, NULL, NULL)) {
                gtk_tree_selection_unselect_all(window.select);
                gtk_tree_selection_select_path(window.select, path);
                gtk_tree_path_free(path);
            }
        }
        
        show_popup_menu(treeview, event, data);
        return TRUE;
    }
    return FALSE;
}

static gboolean on_popup_menu(GtkWidget *treeview, gpointer data) {
    show_popup_menu(treeview, NULL, data);
    return TRUE;
}

void show_text(entry_t *entry) {
    GtkTextIter iter;
    gtk_text_buffer_set_text(window.buffer, "", -1);
    if( entry->crc32 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("CRC32: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->crc32, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->md4 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("MD4: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->md4, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->md5 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("MD5: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->md5, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->sha1 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("SHA1: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->sha1, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->sha256 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("SHA256: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->sha256, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->sha384 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("SHA384: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->sha384, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->sha512 != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("SHA512: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->sha512, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->ed2k != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("ED2K: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->ed2k, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->ed2klink != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("ED2K Link: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->ed2klink, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
    if( entry->tiger != NULL ) {
        (void)gtk_text_buffer_get_end_iter(window.buffer, &iter);
        (void)gtk_text_buffer_insert_with_tags(window.buffer, &iter, _("Tiger: "), -1, window.bold, window.blue, NULL);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, (gchar*)entry->tiger, -1);
        (void)gtk_text_buffer_insert_at_cursor(window.buffer, "\n", -1);
    }
}

void show_text_wrap(entry_t *entry) {
    gdk_threads_enter();
    show_text(entry);

    if( queue->first_entry == NULL ) {
        gtk_widget_set_sensitive(window.abort_button, FALSE);
    }

    gdk_threads_leave();
}

static void selection_text(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data) {
    entry_t *entry = NULL;
    gtk_tree_model_get(model, iter, ENTRY_COLUMN, &entry, -1);
    show_text(entry);
    
    if( entry->crc32 != NULL ) {
        gtk_widget_set_sensitive(window.crc32_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.crc32_button, TRUE);
    }
    if( entry->md4 != NULL ) {
        gtk_widget_set_sensitive(window.md4_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.md4_button, TRUE);
    }
    if( entry->md5 != NULL ) {
        gtk_widget_set_sensitive(window.md5_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.md5_button, TRUE);
    }
    if( entry->sha1 != NULL ) {
        gtk_widget_set_sensitive(window.sha1_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.sha1_button, TRUE);
    }
    if( entry->sha256 != NULL ) {
        gtk_widget_set_sensitive(window.sha256_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.sha256_button, TRUE);
    }
    if( entry->sha384 != NULL ) {
        gtk_widget_set_sensitive(window.sha384_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.sha384_button, TRUE);
    }
    if( entry->sha512 != NULL ) {
        gtk_widget_set_sensitive(window.sha512_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.sha512_button, TRUE);
    }
    if( entry->ed2k != NULL ) {
        gtk_widget_set_sensitive(window.ed2k_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.ed2k_button, TRUE);
    }
    if( entry->ed2klink != NULL ) {
        gtk_widget_set_sensitive(window.ed2klink_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.ed2klink_button, TRUE);
    }
    if( entry->tiger != NULL ) {
        gtk_widget_set_sensitive(window.tiger_button, FALSE);
    } else {
        gtk_widget_set_sensitive(window.tiger_button, TRUE);
    }
}

static void selection_changed(GtkTreeSelection *selection, gpointer data) {
    gtk_tree_selection_selected_foreach(window.select, selection_text, NULL);
}

static void on_add_file(GtkWidget *widget, gpointer data) {
    GtkWidget *dialog;
    uint8_t *filename = NULL;
    
    dialog = gtk_file_chooser_dialog_new(_("Add File"), GTK_WINDOW(window.window), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

    if( status->path )
        gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), (gchar*)status->path);

    if( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
        filename = (uint8_t*)gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    }

    gtk_widget_destroy(dialog);
    
    if( filename != NULL ) {
        uint8_t *ext = get_ext(filename);
        
        if( status->path != NULL ) {
            free(status->path);
            status->path = NULL;
        }
        status->path = get_path(filename);

        if( (ext != NULL) && (strncmp((char*)ext, "sfv", 3) == 0) ) {
            uint8_t *path = get_path(filename);
            uint32_t length = strlen((char*)path);

            char *data = (char*)read_file(filename);

            char *pch = strsep(&data, "\n");
            while( pch != NULL  ) {
                if( pch[0] == ';' )
                    goto end;
                if( pch[0] == ' ' )
                    goto end;

                entry_t *entry = (entry_t*)malloc(sizeof(entry_t));
                memset(entry, 0, sizeof(entry_t));

                char *sch = strtok(pch, " ");

                if( sch != NULL ) {
                    uint32_t len = length + strlen(sch) + 3;
                    entry->path = (uint8_t*)malloc(sizeof(uint8_t)*len);
                    memset(entry->path, 0, len);

                    strcpy((char*)entry->path, (char*)path);
                    strcpy((char*)(entry->path+len), PATH_DELIM);
                    strcpy((char*)(entry->path+len+1), sch);

                    len = strlen(sch)+1;
                    entry->filename = (uint8_t*)malloc(sizeof(uint8_t)*len);
                    memset(entry->filename, 0, len);

                    strcpy((char*)entry->filename, sch);
    
                    sch = strtok(NULL, " ");
                    
                    if( sch != NULL ) {
                        
                        len = strlen(sch)+1;
                        entry->given_crc32 = (uint8_t*)malloc(sizeof(uint8_t)*len);
                        memset(entry->given_crc32, 0, len);

                        strcpy((char*)entry->given_crc32, sch);
                        
                        entry->what = 1;
                        queue_put(queue, entry);
 
                        GtkTreeIter iter;
                        gtk_list_store_append(window.file_store, &iter);
                        gtk_list_store_set(window.file_store, &iter,
                                                PIXBUF_COLUMN, window.grey,
                                                FILE_COLUMN, entry->filename,
                                                ENTRY_COLUMN, entry,
                                            -1);
                    } else {
                        entry_free(entry);
                    }
                } else {
                    entry_free(entry);
                }

                end:
                pch = strsep(&data, "\n");
            }
            
            free(data);
            free(path);

        } else {
            entry_t *entry = (entry_t*)malloc(sizeof(entry_t));
            memset(entry, 0, sizeof(entry_t));

            uint32_t len = strlen((char*)filename)+1;
            
            entry->path = (uint8_t*)malloc(sizeof(uint8_t)*len);
            memset(entry->path, 0, len);

            strcpy((char*)entry->path, (char*)filename);

            entry->filename = get_filename(filename);

            if( status->flags & STATUS_AUTO_CRC32 ) {
                entry->what = 1;
                queue_put(queue, entry);
            }
        
            GtkTreeIter iter;
            gtk_list_store_append(window.file_store, &iter);
            gtk_list_store_set(window.file_store, &iter,
                                    PIXBUF_COLUMN, window.grey,
                                    FILE_COLUMN, entry->filename,
                                    ENTRY_COLUMN, entry,
                                -1);
        }

        free(ext);
        free(filename);
        
        gtk_widget_set_sensitive(window.abort_button, TRUE);
    }
}

static void on_save_sfv(GtkWidget *widget, gpointer data) {
    uint8_t *filename = NULL;

    GtkWidget *dialog;

    dialog = gtk_file_chooser_dialog_new(_("Save SFV"), GTK_WINDOW(window.window), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
    gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);
    
    if( status->path != NULL )
        gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), (gchar*)status->path);

    if( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
        filename = (uint8_t*)gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    }

    gtk_widget_destroy(dialog);

    if( filename != NULL ) {
        
        uint8_t mode[] = "wb";
        file_t *file = file_open(filename, mode);

        if( file ) {
            file_write_line(file, (uint8_t*)"; Generated by "PROGRAM" "VERSION"\n");
            file_write_line(file, (uint8_t*)";\n");

            GtkTreeIter iter;
            gboolean valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(window.file_store), &iter);
        
            while( valid ) {
                entry_t *entry = NULL;
                gtk_tree_model_get(GTK_TREE_MODEL(window.file_store), &iter, ENTRY_COLUMN, &entry, -1);
            
                uint32_t len = strlen((char*)entry->filename)+10;
                uint8_t *buf = (uint8_t*)malloc(sizeof(uint8_t)*len);
                memset(buf, 0, sizeof(uint8_t)*len);
        
                strcat((char*)buf, (char*)entry->filename);
                strcat((char*)buf, " ");
                strcat((char*)buf, (char*)entry->crc32);
                strcat((char*)buf, "\n");
    
                file_write_line(file, buf);

                free(buf);
    
                valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(window.file_store), &iter);
            }

            file_close(file);
        }

        free(filename);
    }
}

static void selection_put(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data) {
    entry_t *entry = NULL;
    gtk_tree_model_get(model, iter, ENTRY_COLUMN, &entry, -1);

    if( entry->crc32 ) {
        uint32_t len = strlen((char*)entry->path)+11;
        uint8_t *to = (uint8_t*)malloc(sizeof(uint8_t)*len);
        memset(to, 0, len);
        uint8_t *ext = NULL;
    
        uint8_t *f = entry->path;
        for( ; *f != '\0'; f=f+1 );
        for( ; f != entry->path; f=f-1 ) {
            if( *f == '.' )
                break;
        }
        
        if( f != entry->path ) {
            ext = f;
            strncpy((char*)to, (char*)entry->path, f - entry->path);
        } else {
            strcpy((char*)to, (char*)entry->path);
        }

        strcat((char*)to, "[");
        strcat((char*)to, (char*)entry->crc32);
        strcat((char*)to, "]");
        
        if( ext != NULL ) {
            strcat((char*)to, (char*)ext);
        }

        file_rename(entry->path, to);

        free(entry->path);
        entry->path = to;

        gtk_list_store_set(window.file_store, iter,
                            PIXBUF_COLUMN, window.green,
                            STATUS_COLUMN, "Found CRC32",
                            -1);
    }
}

static void on_put_crc32(GtkWidget *widget, gpointer data) {
    gtk_tree_selection_selected_foreach(window.select, selection_put, NULL);
}

static void selection_func(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data) {
    entry_t *entry = NULL;
    gtk_tree_model_get(model, iter, ENTRY_COLUMN, &entry, -1);
    
    uint8_t *i = (uint8_t*)data;
    entry->what = *i;
    queue_put(queue, entry);

    show_text(entry);
}

static void on_crc32(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 1;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.crc32_button, FALSE);
}

static void on_md4(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 2;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.md4_button, FALSE);
}

static void on_md5(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 3;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.md5_button, FALSE);
}

static void on_sha1(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 4;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.sha1_button, FALSE);
}

static void on_sha256(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 5;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.sha256_button, FALSE);
}

static void on_sha384(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 6;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.sha384_button, FALSE);
}

static void on_sha512(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 7;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.sha512_button, FALSE);
}

static void on_ed2k(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 8;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.ed2k_button, FALSE);
}

static void on_ed2klink(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 9;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.ed2k_button, FALSE);
    gtk_widget_set_sensitive(window.ed2klink_button, FALSE);
}

static void on_tiger(GtkWidget *widget, gpointer data) {
    gtk_widget_set_sensitive(window.abort_button, TRUE);
    uint8_t i = 10;
    gtk_tree_selection_selected_foreach(window.select, selection_func, &i);
    gtk_widget_set_sensitive(window.tiger_button, FALSE);
}

void check_crc_in_table(entry_t *entry) {
    GtkTreeIter iter;
    gboolean valid;

    gdk_threads_enter();
    valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(window.file_store), &iter);

    while( valid ) {
        entry_t *entry_in_table = NULL;
        gtk_tree_model_get(GTK_TREE_MODEL(window.file_store), &iter, ENTRY_COLUMN, &entry_in_table, -1);

        if( entry == entry_in_table ) {
            if( entry->given_crc32 != NULL ) {
                char *pch = strstr((char*)entry->crc32, (char*)entry->given_crc32);
                if( pch != NULL ) {
                    gtk_list_store_set(window.file_store, &iter,
                                            PIXBUF_COLUMN, window.green,
                                            STATUS_COLUMN, "Found CRC32",
                                            CRC32_COLUMN, entry->crc32,
                                        -1);
                } else {
                    gtk_list_store_set(window.file_store, &iter,
                                            PIXBUF_COLUMN, window.red,
                                            STATUS_COLUMN, "CRC32 not found",
                                            CRC32_COLUMN, entry->crc32,
                                        -1);
                }
            } else {
                if( entry->crc32 ) {
                    char *pch = strstr((char*)entry->filename, (char*)entry->crc32);
                    
                    if( pch != NULL ) {
                        gtk_list_store_set(window.file_store, &iter,
                                                PIXBUF_COLUMN, window.green,
                                                STATUS_COLUMN, "Found CRC32",
                                                CRC32_COLUMN, entry->crc32,
                                            -1);
                    } else {
                        gtk_list_store_set(window.file_store, &iter,
                                                PIXBUF_COLUMN, window.grey,
                                                STATUS_COLUMN, "No CRC32 found",
                                                CRC32_COLUMN, entry->crc32,
                                            -1);
                    }
                }
            }
        }
        valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(window.file_store), &iter);
    }

    gdk_threads_leave();
}

